/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/files/file.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_DOCCOMPILER 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_DOCCOMPILER, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001-2004.  All rights reserved.",
   "January 2004",
   "DocCompiler"
};

static struct KernelBase     *KernelBase = 0;
static struct StringsBase    *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR DocCompilerClass = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR modFileSystem = 0;

/*****************************************************************************
** Class definition.
*/

#define FTYPE_FUNCTION 1
#define FTYPE_METHOD   2
#define FTYPE_ACTION   3
#define FTYPE_EVENT    4

struct DocCompiler {
   OBJECT_HEADER        /* Class header */
   OBJECTID OutputID;   /* The object to output the data to */
   LONG     Static;     /* Set to TRUE if we are static */
   LONG     Clear;      /* Set to TRUE if the output object should be cleared */
   STRING   Destination;

   PRIVATE_FIELDS
   STRING Location;   /* Location of the document directory or specific file */
   STRING FileBuffer;
   STRING FileOutput;
   STRING Description;
   STRING TechnicalOverview;
   STRING Notes;
   LONG FileBufferSize;
   WORD TotalFunctions;
   WORD TotalFields;
   WORD DocType;
   UBYTE MH_Name[40];
   UBYTE MH_Author[80];
   UBYTE MH_Copyright[160];
   UBYTE MH_Version[10];
   UBYTE MH_Date[30];
   UBYTE MH_Keywords[160];
   UBYTE MH_Classes[160];
   UBYTE MH_Short[240];
   UBYTE MH_ID[30];
   UBYTE MH_Status[40];
   UBYTE MH_Category[30];
   UBYTE MH_Include[40];
   UBYTE MH_Type[30];
   UBYTE MH_File[120];
   struct DocFunction * FunctionArray[120];
   struct DocField * Fields[80];
};

#define DOC_MODULE   1
#define DOC_CLASS    2
#define DOC_DOCUMENT 3

static ERROR GET_Location(struct DocCompiler *, STRING *);

static ERROR SET_Destination(struct DocCompiler *, STRING);
static ERROR SET_Location(struct DocCompiler *, STRING);

static struct FieldArray DocCompilerFields[] = {
   { "Output",      0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Static",      0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Clear",       0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Destination", 0, FDF_STRING|FDF_RW,   0, NULL, SET_Destination },
   /*** Virtual fields ***/
   { "Location", 0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Src",      0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   END_FIELD
};

static ERROR COMPILER_Activate(struct DocCompiler *, APTR);
static ERROR COMPILER_ClosingTag(struct DocCompiler *, APTR);
static ERROR COMPILER_Free(struct DocCompiler *, APTR);

static struct ActionArray DocCompilerActions[] = {
   { AC_Activate,   COMPILER_Activate },
   { AC_ClosingTag, COMPILER_ClosingTag },
   { AC_Free,       COMPILER_Free },
   { NULL, NULL }
};

struct DocFunction {
   WORD Number;
   BYTE Name[40];
   BYTE Synonym[40];
   BYTE Synopsis[200];
   BYTE Short[240];
   STRING Description;
   STRING Notes;
   STRING Example;
   WORD TotalInput;
   WORD TotalResults;
   WORD Type;
   FLOAT Version;
   struct DocInput   *Input;
   struct DocResult  *Results;
   struct DocSeeAlso *SeeAlso;
};

struct DocField {
   WORD Number;
   BYTE Name[20];
   BYTE Type[50];
   BYTE Prefix[10];
   BYTE Permissions[30];
   BYTE Short[100];
   BYTE Synonyms[40];
   FLOAT Version;
   STRING Description;
   STRING Notes;
   struct DocSeeAlso *SeeAlso;
};

#define LEN_ERROR     30
#define LEN_ARGUMENT  30
#define LEN_REFERENCE 30
#define LEN_LINK      80
#define LEN_ACTION    20

struct DocInput {
   BYTE Argument[LEN_ARGUMENT];
   STRING Description;
};

struct DocResult {
   BYTE ErrorCode[LEN_ERROR];
   STRING Description;
};

struct DocSeeAlso {
   BYTE Reference[LEN_REFERENCE];
   BYTE Link[LEN_LINK];
};

static void CleanString(STRING);
static void FreeAllocations(struct DocCompiler *);
static void OutputText(struct DocCompiler *, STRING);
static void SortFunctions(struct DocCompiler *);
static void SortFields(struct DocCompiler *);
static ERROR  ProcessFile(struct DocCompiler *);
static ERROR  RecognisedFile(STRING);
static LONG   CountTabulations(STRING);
static ERROR  ExtractTabulation(STRING, STRING, LONG, STRING *);
static STRING ExtractParagraph(STRING *);
static STRING ExtractValue(STRING, STRING, LONG);
STRING WriteFunction(struct DocCompiler *, struct DocFunction *, STRING);

static STRING WriteHTMLFields(struct DocCompiler *, STRING);
static STRING WriteHTMLHeader(struct DocCompiler *, STRING);
static STRING WriteHTMLIntroduction(struct DocCompiler *, STRING);
static STRING WriteHTMLFunctionTable(struct DocCompiler *, STRING);
static STRING WriteHTMLFunctions(struct DocCompiler *, STRING);
static STRING WriteHTMLTail(struct DocCompiler *, STRING);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&DocCompilerClass, NULL,
      FID_BaseClassID|TLONG, ID_DOCCOMPILER,
      FID_Version|TFLOAT,    VER_DOCCOMPILER,
      FID_Name|TSTRING,      "DocCompiler",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      DocCompilerActions,
      FID_Fields|TPTR,       DocCompilerFields,
      FID_Size|TLONG,        sizeof(struct DocCompiler),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modStrings)       { acFree(modStrings);       modStrings = NULL; }
   if (modFileSystem)    { acFree(modFileSystem);    modFileSystem = NULL; }
   if (DocCompilerClass) { acFree(DocCompilerClass); DocCompilerClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name: Activate

###DESCRIPTION###
<p>Activating a DocCompiler object results in the file data being sent to the
object specified in the Output field.</p>

###END###

*****************************************************************************/

static ERROR COMPILER_Activate(struct DocCompiler *Self, APTR Void)
{
   struct acDataChannel Channel;
   struct acWrite write;
   OBJECTPTR file;
   LONG i;

   if (!Self->Location) {
      DPrintF("!Activate:","[DocCompiler:%d] The Location field has not been set.", Self->Head.UniqueID);
      OutputText(Self, "You need to give me a file or directory location to process.");
      return(ERR_FieldNotSet);
   }

   DPrintF("Activate()","[DocCompiler] Location: %s", Self->Location);

   if (Self->OutputID) ActionMsg(AC_Clear, Self->OutputID, NULL);

   if (ProcessFile(Self) IS ERR_Okay) {
      if (Self->OutputID) {
         Channel.ObjectID     = GetUniqueID(Self);
         Channel.DataType     = DATA_TEXT;
         Channel.Version      = 1;
         Channel.Buffer       = Self->FileOutput;
         Channel.Size         = StrLength(Self->FileOutput) + 1;
         Channel.TotalEntries = 1;
         ActionMsg(AC_DataChannel, Self->OutputID, &Channel);
      }

      if ((!Self->Destination) AND (StrLength(Self->MH_Name) > 0)) {
         #define BUFFER_LEN 256
         if (AllocMemory(BUFFER_LEN, MEM_STRING, (APTR *)&Self->Destination, NULL) IS ERR_Okay) {
            if (Self->DocType IS DOC_CLASS) i = StrCopy("pandora:sdk/manual/classes/", Self->Destination, COPY_ALL);
            else if (Self->DocType IS DOC_MODULE) i = StrCopy("pandora:sdk/manual/modules/", Self->Destination, COPY_ALL);
            else i = StrCopy("pandora:sdk/manual/", Self->Destination, COPY_ALL);
            i += StrCopy(Self->MH_Name, Self->Destination + i, BUFFER_LEN - i - 1);
            StrCopy(".html", Self->Destination + i, BUFFER_LEN - i - 1);

            for (i=0; Self->Destination[i]; i++) {
               if (Self->Destination[i] IS ' ') Self->Destination[i] = '_';
            }
            StrLower(Self->Destination);
         }
      }

      if (Self->Destination) {
         if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            FID_Location|TSTRING, Self->Destination,
            TAGEND) IS ERR_Okay) {
            write.Buffer = Self->FileOutput;
            write.Length = StrLength(Self->FileOutput);
            Action(AC_Write, file, &write);
            acFree(file);
         }
      }

      FreeAllocations(Self);
      return(ERR_Okay);
   }

   FreeAllocations(Self);
   return(ERR_Failed);
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR COMPILER_ClosingTag(struct DocCompiler *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      acActivate(Self);
      acFree(Self);
      return(ERR_Okay);
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** DocCompiler: Free
*/

static ERROR COMPILER_Free(struct DocCompiler *Self, APTR Void)
{
   FreeAllocations(Self);

   if (Self->Location) {
      FreeMemory(Self->Location);
      Self->Location = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: FreeAllocations()
*/

void FreeAllocations(struct DocCompiler *Self)
{
   LONG i, j;

   DPrintF("~3DocCompiler:","Freeing allocations.");

   for (i=0; i < Self->TotalFunctions; i++) {
      if (Self->FunctionArray[i]->Description) {
         FreeMemory(Self->FunctionArray[i]->Description);
      }

      if (Self->FunctionArray[i]->Notes) {
         FreeMemory(Self->FunctionArray[i]->Notes);
      }

      if (Self->FunctionArray[i]->Example) {
         FreeMemory(Self->FunctionArray[i]->Example);
      }

      if (Self->FunctionArray[i]->SeeAlso) {
         FreeMemory(Self->FunctionArray[i]->SeeAlso);
      }

      if (Self->FunctionArray[i]->Input) {
         for (j=0; j < Self->FunctionArray[i]->TotalInput; j++) {
            if (Self->FunctionArray[i]->Input[j].Description) {
               FreeMemory(Self->FunctionArray[i]->Input[j].Description);
            }
         }
         FreeMemory(Self->FunctionArray[i]->Input);
      }

      if (Self->FunctionArray[i]->Results) {
         for (j=0; j < Self->FunctionArray[i]->TotalResults; j++) {
            if (Self->FunctionArray[i]->Results[j].Description) {
               FreeMemory(Self->FunctionArray[i]->Results[j].Description);
            }
         }
         FreeMemory(Self->FunctionArray[i]->Results);
      }

      FreeMemory(Self->FunctionArray[i]);
   }

   Self->TotalFunctions = 0;

   for (i=0; i < Self->TotalFields; i++) {
      if (Self->Fields[i]->Description) FreeMemory(Self->Fields[i]->Description);
      if (Self->Fields[i]->Notes) FreeMemory(Self->Fields[i]->Notes);
      FreeMemory(Self->Fields[i]);
   }

   Self->TotalFields = 0;

   if (Self->TechnicalOverview) {
      FreeMemory(Self->TechnicalOverview);
      Self->TechnicalOverview = NULL;
   }

   if (Self->Description) {
      FreeMemory(Self->Description);
      Self->Description = NULL;
   }

   if (Self->Notes) {
      FreeMemory(Self->Notes);
      Self->Notes = NULL;
   }

   if (Self->Destination) {
      FreeMemory(Self->Destination);
      Self->Destination = NULL;
   }

   if (Self->FileBuffer) {
      FreeMemory(Self->FileBuffer);
      Self->FileBuffer = NULL;
   }

   if (Self->FileOutput) {
      FreeMemory(Self->FileOutput);
      Self->FileOutput = NULL;
   }

   Self->FileBufferSize = NULL;

   StepBack();
}

/****************************************************************************
** FIELD: Destination
*/

static ERROR SET_Destination(struct DocCompiler *Self, STRING Value)
{
   LONG i;

   if (Self->Destination)  { FreeMemory(Self->Destination); Self->Destination = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Destination, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Destination[i] = Value[i];
         Self->Destination[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/****************************************************************************
** FIELD: Location
*/

static ERROR GET_Location(struct DocCompiler *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   *Value = NULL;
   return(ERR_FieldNotSet);
}

static ERROR SET_Location(struct DocCompiler *Self, STRING Value)
{
   LONG i;

   if (Self->Location)  { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Functions: ProcessFile()
*/

static ERROR ProcessFile(struct DocCompiler *Self)
{
   OBJECTPTR File;
   struct DocFunction *Function;
   struct acRead Read;
   STRING Buffer, Description, NewBuffer;
   LONG i, j, Flags, FileSize, Offset, SrcType, pos, length, count, type;
   UBYTE Location[512], ErrorBuffer[60], Definition[20];
   ERROR error;

   if (!Self->Location) return(ERR_Failed);

   /*** Presets ***/

   StrCopy("Unnamed", Self->MH_Name, sizeof(Self->MH_Name));
   StrCopy("Rocklyte Systems", Self->MH_Author, sizeof(Self->MH_Author));
   StrCopy("Rocklyte Systems (c) 2004.  All rights reserved.", Self->MH_Copyright, sizeof(Self->MH_Copyright));
   StrCopy("1.0", Self->MH_Version, sizeof(Self->MH_Version));
   StrCopy("-", Self->MH_Date, sizeof(Self->MH_Date));
   StrCopy("module", Self->MH_Keywords, sizeof(Self->MH_Keywords));
   StrCopy("-", Self->MH_Classes, sizeof(Self->MH_Classes));
   StrCopy("-", Self->MH_Short, sizeof(Self->MH_Short));
   StrCopy("-", Self->MH_Category, sizeof(Self->MH_Category));
   StrCopy("-", Self->MH_ID, sizeof(Self->MH_ID));
   StrCopy("Unspecified", Self->MH_Status, sizeof(Self->MH_Status));
   StrCopy("-", Self->MH_Include, sizeof(Self->MH_Include));
   StrCopy("-", Self->MH_Type, sizeof(Self->MH_Type));

   Self->TotalFunctions = 0;

   /* Open the file and read the data, or if it is a directory, load
   ** all the files into one massive buffer.
   */

   error = AnalyseLocation(Self->Location, &SrcType);

   if (SrcType IS LOC_FILE) {
      if (NewPrivateObject(ID_FILE, NULL, &File) IS ERR_Okay) {
         Flags = FL_READ;
         SetField(File, FID_Flags, FT_LONG, &Flags);
         SetField(File, FID_Location, FT_POINTER, Self->Location);
         if (Action(AC_Init, File, NULL) IS ERR_Okay) {
            GetField(File, FID_Size, FT_LONG, &Read.Length);
            if (AllocMemory(Read.Length+1, MEM_STRING|MEM_NOCLEAR, (APTR *)&Self->FileBuffer, NULL) IS ERR_Okay) {
               Read.Buffer = Self->FileBuffer;
               if (Action(AC_Read, File, &Read) IS ERR_Okay) {
                  ((BYTE *)Read.Buffer)[Read.Length] = 0; /* Terminate the end of file */
                  Self->FileBufferSize = Read.Length;
               }
               else error = ObjectError(Self, ERH_Activate, ERR_Read);
            }
            else error = ObjectError(Self, ERH_Activate, ERR_Memory);
         }
         else error = ObjectError(Self, ERH_Activate, ERR_Init);

         Action(AC_Free, File, NULL);
      }
      else error = ObjectError(Self, ERH_Activate, ERR_NewObject);
   }
   else if (SrcType IS LOC_DIRECTORY) {
      struct FileInfo *list, *scan;
      if (ReadDirectory(Self->Location, RDF_FILE, &list) IS ERR_Okay) {

         /*** Load each file into the memory buffer ***/

         for (scan=list; scan; scan=scan->Next) {
            if (RecognisedFile(scan->Name) IS ERR_True) {
               j = StrCopy(Self->Location, Location, COPY_ALL);
               if (Location[j-1] != '/') Location[j++] = '/';
               StrCopy(scan->Name, Location+j, COPY_ALL);

               if (CreateObject(ID_FILE, NF_CHILD, &File, NULL,
                  FID_Location|TSTRING, Location,
                  FID_Flags|TLONG, FL_READ,
                  TAGEND) IS ERR_Okay) {

                  GetField(File, FID_Size, FT_LONG, &FileSize);

                  if (!Self->FileBuffer) {
                     if (AllocMemory(FileSize, MEM_STRING, (APTR *)&Self->FileBuffer, NULL) IS ERR_Okay) {
                        Self->FileBufferSize = FileSize;
                        Offset = 0;
                     }
                     else break;
                  }
                  else {
                     if (ReallocMemory(Self->FileBuffer, Self->FileBufferSize + FileSize, (void **)&Self->FileBuffer, NULL) IS ERR_Okay) {
                        Offset = Self->FileBufferSize;
                        Self->FileBufferSize += FileSize;
                     }
                     else break;
                  }

                  /*** Read the file contents ***/
                  Read.Buffer = Self->FileBuffer + Offset;
                  Read.Length = FileSize;
                  Action(AC_Read, File, &Read);

                  Action(AC_Free, File, NULL);
               }
            }
         }
         FreeMemory(list);
      }
   }
   else {
      DPrintF("DocCompiler:","Location \"%s\" is invalid.", Self->Location);
      OutputText(Self, "The location that you asked me to use is invalid.");
      error = ERR_Failed;
   }

   if (Self->FileBufferSize <= 0) error = ERR_Failed;

   if (error IS ERR_Okay) {
      DPrintF("DocCompiler:","The load operation is complete (%d bytes loaded) - now stripping the data.", Self->FileBufferSize);

      /*** Strip out everything that isn't encapsulated by the hash ### marks ***/

      if ((error = AllocMemory(Self->FileBufferSize, NULL, (APTR *)&NewBuffer, NULL)) IS ERR_Okay) {
         i = 0;
         j = 0;
         while (Self->FileBuffer[i] != 0) {
            if (StrCompare("###", Self->FileBuffer+i, 3, STR_CASE) IS ERR_Okay) {
               pos = StrSearch("###END###", Self->FileBuffer+i, STR_CASE);
               if (pos != -1) {
                  pos += StrLength("###END###");
                  while (pos > 0) {
                     NewBuffer[j++] = Self->FileBuffer[i++];
                     pos--;
                  }
                  NewBuffer[j++] = 10;
               }
               else i++;
            }
            else i++;
         }

         NewBuffer[j++] = 0;
         FreeMemory(Self->FileBuffer);
         Self->FileBuffer = NewBuffer;
         Self->FileBufferSize = j;
         DPrintF("DocCompiler:","Strip complete.");
      }
   }

   /*** Search for module and class headers ***/

   if (error IS ERR_Okay) {
      Buffer = Self->FileBuffer;

      if ((pos = StrSearch("###MODULE###", Buffer, FALSE)) != -1) Self->DocType = DOC_MODULE;
      else if ((pos = StrSearch("###CLASS###", Buffer, FALSE)) != -1) Self->DocType = DOC_CLASS;
      else if ((pos = StrSearch("###DOCUMENT###", Buffer, FALSE)) != -1) Self->DocType = DOC_DOCUMENT;
      else {
         OutputText(Self, "The source does not contain ###MODULE### or ###CLASS### markers.");
         error = ERR_Search;
         goto exit;
      }

      Buffer = StrNextLine(Buffer + pos);
      if ((count = CountTabulations(Buffer))) {
         while ((Buffer) AND (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay)) {
            for (i=0; (Buffer[i] != '\n') AND (Buffer[i] != ' '); i++);
            if ((i > 2) AND (Buffer[i-1] IS ':')) {
               if (ExtractTabulation(Buffer, Definition, sizeof(Definition), &Description) IS ERR_Okay) {
                  if (Description) {
                     if (StrCompare("Author", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Author, sizeof(Self->MH_Author));
                     }
                     else if (StrCompare("Category", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Category, sizeof(Self->MH_Category));
                     }
                     else if (StrCompare("Classes", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Classes, sizeof(Self->MH_Classes));
                     }
                     else if (StrCompare("Copyright", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Copyright, sizeof(Self->MH_Copyright));
                     }
                     else if (StrCompare("Date", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Date, sizeof(Self->MH_Date));
                     }
                     else if (StrCompare("ID", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_ID, sizeof(Self->MH_ID));
                     }
                     else if (StrCompare("Status", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Status, sizeof(Self->MH_Status));
                     }
                     else if (StrCompare("Include", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Include, sizeof(Self->MH_Include));
                     }
                     else if (StrCompare("Keywords", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Keywords, sizeof(Self->MH_Keywords));
                     }
                     else if (StrCompare("Location", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (!Self->Destination) {
                           Self->Destination = Description;
                           Description = NULL;
                        }
                     }
                     else if (StrCompare("Name", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Name, sizeof(Self->MH_Name));
                     }
                     else if (StrCompare("Short", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Short, sizeof(Self->MH_Short));
                     }
                     else if (StrCompare("Type", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (StrCompare("Class", Description, NULL, STR_MATCHLENGTH) IS ERR_Okay) Self->DocType = DOC_CLASS;
                        else if (StrCompare("Module", Description, NULL, STR_MATCHLENGTH) IS ERR_Okay) Self->DocType = DOC_MODULE;
                        else if (StrCompare("Document", Description, NULL, STR_MATCHLENGTH) IS ERR_Okay) Self->DocType = DOC_DOCUMENT;
                        else DPrintF("!DocCompiler:","Unsupported type \"%s\".", Description);
                     }
                     else if (StrCompare("Version", Definition, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(Description, Self->MH_Version, sizeof(Self->MH_Version));
                     }
                     else DPrintF("!DocCompiler:","Unsupported definition \"%s\".", Definition);

                     if (Description) FreeMemory(Description);
                  }
               }
            }
            Buffer = StrNextLine(Buffer);
         }
      }

      if (StrCompare("###DESCRIPTION", Buffer, 14, NULL) IS ERR_Okay) {
         Buffer = StrNextLine(Buffer);
         Self->Description = ExtractParagraph(&Buffer);
      }

      if (StrCompare("###NOTE", Buffer, 7, NULL) IS ERR_Okay) {
         Buffer = StrNextLine(Buffer);
         Self->Notes = ExtractParagraph(&Buffer);
      }

      if (StrCompare("###TECHNICAL", Buffer, 12, NULL) IS ERR_Okay) {
         Buffer = StrNextLine(Buffer);
         Self->TechnicalOverview = ExtractParagraph(&Buffer);
      }

      if (StrCompare("###ACTIONS", Buffer, 8, NULL) IS ERR_Okay) {
         Buffer = StrNextLine(Buffer);

         while ((Buffer) AND (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay)) {
            for (i=0; (Buffer[i] != '\n') AND (Buffer[i] != ' '); i++);
            if ((i > 2) AND (Buffer[i-1] IS ':')) {
               if (AllocMemory(sizeof(struct DocFunction), NULL, (APTR *)&Function, NULL) IS ERR_Okay) {
                  if (ExtractTabulation(Buffer, Function->Name, LEN_ACTION, &Description) IS ERR_Okay) {
                     Self->FunctionArray[Self->TotalFunctions] = Function;
                     Function->Type = FTYPE_ACTION;
                     StrCopy(Description, Function->Short, sizeof(Function->Short));
                     Self->TotalFunctions++;
                     FreeMemory(Description);
                  }
                  else FreeMemory(Function);
               }
            }
            Buffer = StrNextLine(Buffer);
         }
      }
   }

   /*** Search and process functions, methods and actions ***/

   if (error IS ERR_Okay) {
      Buffer = Self->FileBuffer;
      while (Buffer) {
         /*** Find the next action/method/function/event in the buffer ***/

         if ((pos = StrSearch("###", Buffer, FALSE)) != -1) {
            if (StrCompare("###FUNCTION###", Buffer+pos, NULL, NULL) IS ERR_Okay)    type = FTYPE_FUNCTION;
            else if (StrCompare("###METHOD###", Buffer+pos, NULL, NULL) IS ERR_Okay) type = FTYPE_METHOD;
            else if (StrCompare("###ACTION###", Buffer+pos, NULL, NULL) IS ERR_Okay) type = FTYPE_ACTION;
            else if (StrCompare("###EVENT###", Buffer+pos, NULL, NULL) IS ERR_Okay)  type = FTYPE_EVENT;
            else type = NULL;
         }
         else break;

         Buffer = StrNextLine(Buffer + pos);

         /*** Read each function definition line ***/

         if ((type) AND (AllocMemory(sizeof(struct DocFunction), NULL, (APTR *)&Function, NULL) IS ERR_Okay)) {
            StrCopy("Undefined", Function->Name, COPY_ALL);
            Function->Synopsis[0] = 0;
            StrCopy("...", Function->Short, COPY_ALL);
            Function->Type = type;

            while (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay) {
               if ((StrCompare(Buffer, "Name:", 5, NULL) IS ERR_Okay) OR
                   (StrCompare(Buffer, "Names:", 6, NULL) IS ERR_Okay)) {
                  ExtractValue(Buffer, Function->Name, sizeof(Function->Name));
                  for (i=0; Function->Name[i] AND (Function->Name[i] != '('); i++);
                  Function->Name[i] = 0;
               }
               else if (StrCompare(Buffer, "Synonym:", 8, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Function->Synonym, sizeof(Function->Synonym));
                  for (i=0; Function->Synonym[i] AND (Function->Synonym[i] != '('); i++);
                  Function->Synonym[i] = 0;
               }
               else if (StrCompare(Buffer, "Synopsis:", 9, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Function->Synopsis, sizeof(Function->Synopsis));
               }
               else if (StrCompare(Buffer, "Arguments:", 10, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Function->Synopsis, sizeof(Function->Synopsis));
               }
               else if (StrCompare(Buffer, "Short:", 6, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Function->Short, sizeof(Function->Short));
               }
               else if (StrCompare(Buffer, "Version", 7, NULL) IS ERR_Okay) {
                  Function->Version = StrToFloat(Buffer);
               }
               else {
                  if ((length = StrLineLength(Buffer)) > 0) {
                     if (length > 59) length = 59;
                     ErrorBuffer[CharCopy(Buffer, ErrorBuffer, length)] = 0;
                     DPrintF("!DocCompiler:","Unsupported definition \"%s\"", ErrorBuffer);
                  }
               }

               Buffer = StrNextLine(Buffer);
               if (!Buffer) break;
            }

            if (StrCompare("###DESCRIPTION", Buffer, 14, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
               Function->Description = ExtractParagraph(&Buffer);
            }

            if (StrCompare("###NOTE", Buffer, 7, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
               Function->Notes = ExtractParagraph(&Buffer);
            }

            if (StrCompare("###EXAMPLE", Buffer, 10, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
               Function->Example = ExtractParagraph(&Buffer);
            }

            if (StrCompare("###INPUT", Buffer, 8, NULL) IS ERR_Okay) {
               LONG count, ino;

               Buffer = StrNextLine(Buffer);
               if ((count = CountTabulations(Buffer))) {
                  if (AllocMemory(sizeof(struct DocInput) * count, NULL, (APTR *)&Function->Input, NULL) IS ERR_Okay) {
                     Function->TotalInput = count;
                     ino = 0;
                     while ((Buffer) AND (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay)) {
                        for (i=0; (Buffer[i] != '\n') AND (Buffer[i] != ' '); i++);
                        if ((i > 2) AND (Buffer[i-1] IS ':')) {
                           if (ExtractTabulation(Buffer, Function->Input[ino].Argument, LEN_ARGUMENT, &Description) IS ERR_Okay) {
                              Function->Input[ino].Description = Description;
                              ino++;
                           }
                        }
                        Buffer = StrNextLine(Buffer);
                     }
                  }
               }
               else DPrintF("!DocCompiler:","No input arguments found in function %s.", Function->Name);
            }

            if (StrCompare("###RESULT", Buffer, 9, NULL) IS ERR_Okay) {
               LONG count, no;

               Buffer = StrNextLine(Buffer);
               if ((count = CountTabulations(Buffer))) {
                  if (AllocMemory(sizeof(struct DocResult) * count, NULL, (APTR *)&Function->Results, NULL) IS ERR_Okay) {
                     Function->TotalResults = count;
                     no = 0;
                     while ((Buffer) AND (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay)) {
                        for (i=0; (Buffer[i] != '\n') AND (Buffer[i] != ' '); i++);
                        if ((i > 2) AND (Buffer[i-1] IS ':')) {
                           if (ExtractTabulation(Buffer, Function->Results[no].ErrorCode, LEN_ERROR, &Description) IS ERR_Okay) {
                              Function->Results[no].Description = Description;
                              no++;
                           }
                        }
                        Buffer = StrNextLine(Buffer);
                     }
                  }
               }
               else if (AllocMemory(sizeof(struct DocResult), NULL, (APTR *)&Function->Results, NULL) IS ERR_Okay) {
                  Function->TotalResults = 1;
                  Function->Results->Description = ExtractParagraph(&Buffer);
                  CleanString(Function->Results->Description);
               }
            }

            if (StrCompare("###SEE ALSO", Buffer, 11, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
            }

            /* If an entry for this function already exists, replace it.
            ** Otherwise add the function to the end of the array.
            */

            for (i=0; i < Self->TotalFunctions; i++) {
               if (StrCompare(Function->Name, Self->FunctionArray[i]->Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  FreeMemory(Self->FunctionArray[i]);
                  Self->FunctionArray[i] = Function;
                  break;
               }
            }

            if (i >= Self->TotalFunctions) Self->FunctionArray[Self->TotalFunctions++] = Function;
         }
      }

      if (Self->TotalFunctions > 0) {
         DPrintF("DocCompiler:","%d functions were found in the document.", Self->TotalFunctions);
         SortFunctions(Self);
      }
   }

   /*** Search and process fields ***/

   if (error IS ERR_Okay) {
      LONG fno;

      Buffer = Self->FileBuffer;
      fno = 0;
      while ((Buffer) AND (pos = StrSearch("###FIELD###", Buffer, FALSE)) != -1) {
         Buffer = StrNextLine(Buffer + pos);

         if (AllocMemory(sizeof(struct DocFunction), NULL, (APTR *)&Self->Fields[fno], NULL) IS ERR_Okay) {
            StrCopy("Undefined", Self->Fields[fno]->Name, COPY_ALL);
            StrCopy("Undefined", Self->Fields[fno]->Type, COPY_ALL);
            StrCopy("Undefined", Self->Fields[fno]->Permissions, COPY_ALL);
            StrCopy("...", Self->Fields[fno]->Short, COPY_ALL);
            Self->Fields[fno]->Prefix[0] = 0;
            Self->Fields[fno]->Synonyms[0] = 0;

            while (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay) {
               if (StrCompare(Buffer, "Name", 4, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Name, sizeof(Self->Fields[0]->Name));
               }
               else if (StrCompare(Buffer, "Type:", 5, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Type, sizeof(Self->Fields[0]->Type));
               }
               else if (StrCompare(Buffer, "Short:", 6, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Short, sizeof(Self->Fields[0]->Short));
               }
               else if (StrCompare(Buffer, "Prefix:", 7, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Prefix, sizeof(Self->Fields[0]->Prefix));
               }
               else if (StrCompare(Buffer, "Status:", 7, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Permissions, sizeof(Self->Fields[0]->Permissions));
               }
               else if (StrCompare(Buffer, "Synonym", 7, NULL) IS ERR_Okay) {
                  ExtractValue(Buffer, Self->Fields[fno]->Synonyms, sizeof(Self->Fields[0]->Synonyms));
               }
               else if (StrCompare(Buffer, "Version", 7, NULL) IS ERR_Okay) {
                  Self->Fields[fno]->Version = StrToFloat(Buffer);
               }
               else {
                  if ((length = StrLineLength(Buffer)) > 0) {
                     if (length > 59) length = 59;
                     ErrorBuffer[CharCopy(Buffer, ErrorBuffer, length)] = 0;
                     DPrintF("!DocCompiler:","Unsupported definition \"%s\"", ErrorBuffer);
                  }
               }

               Buffer = StrNextLine(Buffer);
               if (!Buffer) break;
            }

            if (StrCompare("###DESCRIPTION", Buffer, 14, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
               Self->Fields[fno]->Description = ExtractParagraph(&Buffer);
            }

            if (StrCompare("###NOTE", Buffer, 7, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
               Self->Fields[fno]->Notes = ExtractParagraph(&Buffer);
            }

            if (StrCompare("###SEE ALSO", Buffer, 11, NULL) IS ERR_Okay) {
               Buffer = StrNextLine(Buffer);
            }

            fno++;
         }
       }

      if ((Self->TotalFields = fno) > 0) {
         DPrintF("DocCompiler:","%d fields were found in the document.", Self->TotalFields);
         SortFields(Self);
      }
   }

   /*** Process the FileBuffer into the FileOutput memory buffer ***/

   if (error IS ERR_Okay) {
      if ((error = AllocMemory(512000, MEM_NOCLEAR, (APTR *)&Self->FileOutput, NULL)) IS ERR_Okay) {
         Buffer = WriteHTMLHeader(Self, Self->FileOutput);
         Buffer = WriteHTMLIntroduction(Self, Buffer);
         Buffer = WriteHTMLFunctionTable(Self, Buffer);
         Buffer = WriteHTMLFunctions(Self, Buffer);
         Buffer = WriteHTMLFields(Self, Buffer);
         Buffer = WriteHTMLTail(Self, Buffer);
      }
      else OutputText(Self, "Your system is too low on memory for me to complete the process.");
   }

exit:

   return(error);
}

/****************************************************************************/

static void CleanString(STRING str)
{
   LONG j, i;

   if (!str) return;
   if (!*str) return;

   /*** Convert return codes to spaces ***/
   for (i=0; str[i]; i++) {
      if (str[i] IS '\n') str[i] = ' ';
   }

   /*** Shrink the string by eliminating double spaces ***/

   for (i=1; str[i]; i++) {
      if ((str[i] IS ' ') AND (str[i-1] IS ' ')) {
         for (j=i; str[j]; j++) {
            str[j-1] = str[j];
         }
         str[j-1] = 0;
         i = 1;
      }
   }

   if (str[i-1] IS ' ') str[i-1] = 0;
}

/****************************************************************************/

static ERROR RecognisedFile(STRING Location)
{
   LONG Length;

   Length = StrLength(Location);
   if (Length <= 3) return(ERR_False);
   if ((Location[Length-2] IS '.') AND (Location[Length-1] IS 'c')) return(ERR_True);
   return(ERR_False);
}

/****************************************************************************/

static LONG CountTabulations(STRING Buffer)
{
   LONG count, i;

   count = 0;
   while ((Buffer) AND (StrCompare("###", Buffer, 3, STR_CASE) != ERR_Okay)) {
      for (i=0; (Buffer[i] != '\n') AND (Buffer[i] != ' ') AND (Buffer[i]); i++);
      if ((i > 1) AND (Buffer[i-1] IS ':')) count++;
      Buffer = StrNextLine(Buffer);
   }
   return(count);
}

/****************************************************************************/

static STRING ExtractValue(STRING Source, STRING Destination, LONG DestSize)
{
   LONG i, count;
   STRING Temp;

   while (*Source != ':') Source++;
   if (*Source IS ':') Source++;
   while ((*Source) AND (*Source <= 0x20)) Source++;

   /* Count all the bytes up to the next argument or ### hash, then
   ** copy all of those bytes into the description for the argument.
   */

   count = 0;
   Temp = Source;
   while ((Temp) AND (StrCompare("###", Temp, 3, STR_CASE) != ERR_Okay)) {
      i = 0;
      while ((Temp[i] != '\n') AND (Temp[i] != ' ')) i++;
      if ((i > 2) AND (Temp[i-1] IS ':')) break;
      else count += StrLineLength(Temp) + 1;
      Temp = StrNextLine(Temp);
   }

   if (count > 1) {
      count--; /* (Ignore the trailing return code) */
      if (count < DestSize) Destination[CharCopy(Source, Destination, count)] = 0;
      else StrCopy(Source, Destination, DestSize);
      CleanString(Destination);
      return(Source + count);
   }
   else return(Source);
}

/****************************************************************************/

static ERROR ExtractTabulation(STRING Buffer, STRING Argument, LONG ArgLength, STRING *Description)
{
   LONG i, count;
   STRING Temp;

   *Description = NULL;

   i = 0;
   while (Buffer[i] != ':') i++;
   if (i > ArgLength) i = ArgLength;
   Argument[CharCopy(Buffer, Argument, i)] = 0;
   Buffer += i+1;
   while (*Buffer IS ' ') Buffer++;

   /* Count all the bytes up to the next argument or ### hash, then
   ** copy all of those bytes into the description for the argument.
   */

   count = 0;
   Temp = Buffer;
   while ((Temp) AND (StrCompare("###", Temp, 3, STR_CASE) != ERR_Okay)) {
      i = 0;
      while ((Temp[i] != '\n') AND (Temp[i] != ' ')) i++;
      if ((i > 2) AND (Temp[i-1] IS ':')) break;
      else count += StrLineLength(Temp) + 1;
      Temp = StrNextLine(Temp);
   }

   if (count > 1) {
      count--; /* (Ignore the trailing return code) */
      if (AllocMemory(count+1, MEM_STRING|MEM_NOCLEAR, (APTR *)Description, NULL) IS ERR_Okay) {
         Description[0][CharCopy(Buffer, *Description, count)] = 0;
         CleanString(*Description);
         return(ERR_Okay);
      }
      else return(ERR_Memory);
   }
   else return(ERR_Failed);
}

/****************************************************************************/

static STRING ExtractParagraph(STRING *Buffer)
{
   STRING str;
   LONG count;
   BYTE *Temp;

   Temp = *Buffer;
   count = 0;
   while ((Temp) AND (StrCompare("###", Temp, 3, STR_CASE) != ERR_Okay)) {
      count += StrLineLength(Temp) + 1;
      Temp = StrNextLine(Temp);
   }

   str = NULL;
   if (count > 20) {
      if (AllocMemory(count+1, MEM_STRING|MEM_NOCLEAR, (APTR *)&str, NULL) IS ERR_Okay) {
         /*if (StrCompare("<p>", *Buffer, 3, NULL) IS ERR_Okay) StrCopy((*Buffer)+3, str, count-3);
         else*/ str[CharCopy(*Buffer, str, count)] = 0;
      }
   }

   while ((*Buffer) AND (StrCompare("###", *Buffer, 3, STR_CASE) != ERR_Okay)) {
      *Buffer = StrNextLine(*Buffer);
   }

   return(str);
}

/*****************************************************************************
** Function: SortFunctions()
*/

static LONG CompareStrings(STRING Name1, STRING Name2)
{
   if ((!Name1) OR (!Name2)) return(FALSE);

   while ((*Name1) AND (*Name2)) {
      if (*Name1 < *Name2) {
         return(TRUE);
      }
      else if (*Name1 > *Name2) {
         return(FALSE);
      }

      Name1++;
      Name2++;
   }
   return(FALSE);
}

static void SortFunctions(struct DocCompiler *Self)
{
   register WORD j;
   LONG count, index;

   count = 1;
   index = 0;
   while (index != -1) {
      index = -1;
      for (j=0; j < Self->TotalFunctions; j++) {
         if (!Self->FunctionArray[j]->Number) {
            if (index IS -1) {
               index = j;
            }
            else if (CompareStrings(Self->FunctionArray[j]->Name, Self->FunctionArray[index]->Name) IS TRUE) {
               index = j;
            }
         }
      }

      if (index != -1) Self->FunctionArray[index]->Number = count++;
   }
}

static void SortFields(struct DocCompiler *Self)
{
   register WORD j;
   LONG count, index;

   count = 1;
   index = 0;
   while (index != -1) {
      index = -1;
      for (j=0; j < Self->TotalFields; j++) {
         if (!Self->Fields[j]->Number) {
            if (index IS -1) {
               index = j;
            }
            else if (CompareStrings(Self->Fields[j]->Name, Self->Fields[index]->Name) IS TRUE) {
               index = j;
            }
         }
      }

      if (index != -1) Self->Fields[index]->Number = count++;
   }
}

/*****************************************************************************
** Function: WriteHTMLHeader()
*/

static UBYTE style[] = " <style type=\"text/css\">\n\
<!--\n\
td { font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 10pt; text-align: left; vertical-align: top}\n\
pre { font-size: 10pt ; font-family: \"Courier New\", Courier, mono}\n\
body { font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 10pt}\n\
h1 { font-size: 18pt; font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; color: #000080 }\n\
h2 { font-size: 16pt; font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; color: #000000 }\n\
h3 { font-size: 12pt; font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold }\n\
.copyright { font-family: Arial, Helvetica, sans-serif; font-size: 8pt; font-style: normal; font-weight: normal; color: #000000 }\n\
.tableheader { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold}\n\
.tableentry { font-family: Arial, Helvetica, sans-serif; font-style: normal}\n\
.fh { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; background-color: #f0f0ff;}\n\
.fe { font-style: normal; background-color: #f9f9f9;}\n\
.actionheader { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; background-color: #fff0f0;}\n\
.actionentry { font-style: normal; background-color: #f9f9f9;}\n\
.functionheader { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; background-color: #f0f0ff;}\n\
.functionentry { font-style: normal; background-color: #f9f9f9;}\n\
.methodheader { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; background-color: #e5e5e5;}\n\
.methodentry { font-style: normal; background-color: #f9f9f9;}\n\
.eventheader { font-family: Arial, Helvetica, sans-serif; font-style: normal; font-weight: bold; background-color: #f6f6e0;}\n\
.evententry { font-style: normal; background-color: #f9f9f9;}\n\
-->\n\
 </style>\n";

static STRING WriteHTMLHeader(struct DocCompiler *Self, STRING Buffer)
{
   Buffer += StrFormat(Buffer, NULL, "<html>\n<head>\n <title>%s</title>\n", Self->MH_Name);

   if (StrLength(Self->MH_Author) > 2) Buffer += StrFormat(Buffer, NULL, " <meta name=\"author\" content=\"%s\">\n", Self->MH_Author);
   if (StrLength(Self->MH_Short) > 10) Buffer += StrFormat(Buffer, NULL, " <meta name=\"description\" content=\"%s\">\n", Self->MH_Short);
   if (StrLength(Self->MH_Keywords) > 5) Buffer += StrFormat(Buffer, NULL, " <meta name=\"keywords\" content=\"%s\">\n", Self->MH_Keywords);

   Buffer += StrCopy(style, Buffer, COPY_ALL);
   Buffer += StrCopy("</head>\n\n<body bgcolor=\"#FFFFFF\" text=\"#000000\" vlink=\"#0000FF\" link=\"#0000FF\">\n\n", Buffer, COPY_ALL);

   return(Buffer);
}

/*****************************************************************************
** Function: WriteHTMLIntroduction()
*/

static STRING WriteHTMLIntroduction(struct DocCompiler *Self, STRING Buffer)
{
   WORD pos, i;

   if (Self->DocType IS DOC_CLASS) {
      Buffer += StrCopy("<h1>Class Documentation</h1>\n", Buffer, COPY_ALL);
      Buffer += StrCopy("<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Name:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }
   else {
      Buffer += StrCopy("\n<h1>", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
      Buffer += StrCopy("</h1>\n<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);
   }

   Buffer += StrCopy(" <tr><td class=\"tableheader\">Version:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
   Buffer += StrCopy(Self->MH_Version, Buffer, COPY_ALL);
   Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

   if (Self->MH_ID[0] != '-') {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">ID:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_ID, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   if (Self->MH_Status[0] != '-') {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Status:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Status, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   if (Self->MH_Type[0] != '-') {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Type:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Type, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   if (Self->MH_Category[0] != '-') {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Category:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Category, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   if (Self->MH_Include[0] != '-') {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Include:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Include, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   Buffer += StrCopy(" <tr><td class=\"tableheader\">Date:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
   Buffer += StrCopy(Self->MH_Date, Buffer, COPY_ALL);
   Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

   Buffer += StrCopy(" <tr><td class=\"tableheader\">Author:</td><td class=\"tableentry\">", Buffer, COPY_ALL);
   Buffer += StrCopy(Self->MH_Author, Buffer, COPY_ALL);
   Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

   Buffer += StrCopy(" <tr><td class=\"tableheader\">Copyright:&nbsp;&nbsp;</td><td class=\"tableentry\">", Buffer, COPY_ALL);
   Buffer += StrCopy(Self->MH_Copyright, Buffer, COPY_ALL);
   Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

   if (StrLength(Self->MH_Short) > 10) {
      Buffer += StrCopy(" <tr><td class=\"tableheader\">Short:&nbsp;&nbsp;</td><td class=\"tableentry\">", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Short, Buffer, COPY_ALL);
      Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
   }

   Buffer += StrCopy("</table>\n</font>\n<br>\n<hr color=\"#000000\" size=\"2\"><br>\n", Buffer, COPY_ALL);

   /*** Description ***/

   if (Self->Description) {
      Buffer += StrCopy("\n<h2>Description</h2>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->Description, Buffer, COPY_ALL);
      *Buffer++ = '\n';
   }

   /*** Notes ***/

   if (Self->Notes) {
      Buffer += StrCopy("\n<h2>Note</h2>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->Notes, Buffer, COPY_ALL);
      *Buffer++ = '\n';
   }

   /*** Technical Overview ***/

   if (Self->TechnicalOverview) {
      Buffer += StrCopy("\n<h2>Technical Overview</h2>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->TechnicalOverview, Buffer, COPY_ALL);
      *Buffer++ = '\n';
   }

   /*** Action Index ***/

   for (i=0; i < Self->TotalFunctions; i++) {
      if (Self->FunctionArray[i]->Type IS FTYPE_ACTION) break;
   }

   if (i < Self->TotalFunctions) {
      if (Self->DocType IS DOC_CLASS) {
         Buffer += StrCopy("<h2>Actions</h2>\n\n<p>The ", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
         Buffer += StrCopy(" class supports the following actions:</p>\n\n", Buffer, COPY_ALL);
      }
      else Buffer += StrCopy("<h2>Action Index</h2>\n\n", Buffer, COPY_ALL);

      Buffer += StrCopy("<blockquote>\n<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);

      pos = 1;
      for (i=0; i < Self->TotalFunctions; i++) {
         if (Self->FunctionArray[i]->Number IS pos) {
            if (Self->FunctionArray[i]->Type IS FTYPE_ACTION) {
               if (Self->DocType IS DOC_CLASS) {
                  if (Self->FunctionArray[i]->Description) {
                     Buffer += StrFormat(Buffer, 1024, " <tr><td><i><a href=\"#%s\">%s</a></i>&nbsp;&nbsp;</td><td>%s</td></tr>\n", Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Short);
                  }
                  else Buffer += StrFormat(Buffer, 1024, " <tr><td><a href=\"../general/action_list.html#%s\">%s</a>&nbsp;&nbsp;</td><td>%s</td></tr>\n", Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Short);
               }
               else if (Self->FunctionArray[i]->Description) {
                  Buffer += StrFormat(Buffer, 1024, " <tr><td><a href=\"#%s\">%s</a>&nbsp;&nbsp;</td><td>%s</td></tr>\n", Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Short);
               }
            }
            pos++;
            i = -1;
         }
      }
      Buffer += StrCopy("</td></tr></table></blockquote>\n\n", Buffer, COPY_ALL);
   }

   /*** Method Index ***/

   for (i=0; i < Self->TotalFunctions; i++) {
      if (Self->FunctionArray[i]->Type IS FTYPE_METHOD) break;
   }

   if (i < Self->TotalFunctions) {
      Buffer += StrCopy("<h2>Methods</h2>\n\n<p>The ", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
      Buffer += StrCopy(" class implements the following methods:</p>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy("<blockquote>\n<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);

      pos = 1;
      for (i=0; i < Self->TotalFunctions; i++) {
         if (Self->FunctionArray[i]->Number IS pos) {
            if (Self->FunctionArray[i]->Type IS FTYPE_METHOD) {
               Buffer += StrFormat(Buffer, 1024, " <tr><td><a href=\"#%s()\">%s</a>&nbsp;&nbsp;</td><td>%s</td></tr>\n", Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Short);
            }
            pos++;
            i = -1;
         }
      }
      Buffer += StrCopy("</td></tr></table></blockquote>\n\n", Buffer, COPY_ALL);
   }

   /*** Event Index ***/

   for (i=0; i < Self->TotalFunctions; i++) {
      if (Self->FunctionArray[i]->Type IS FTYPE_EVENT) break;
   }

   if (i < Self->TotalFunctions) {
      if (Self->DocType IS DOC_CLASS) {
         Buffer += StrCopy("<h2>Events</h2>\n\n<p>The ", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
         Buffer += StrCopy(" class supports the following events:</p>\n\n", Buffer, COPY_ALL);
      }
      else Buffer += StrCopy("<h2>Event Index</h2>\n\n", Buffer, COPY_ALL);

      Buffer += StrCopy("<blockquote>\n<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);

      pos = 1;
      for (i=0; i < Self->TotalFunctions; i++) {
         if (Self->FunctionArray[i]->Number IS pos) {
            if (Self->FunctionArray[i]->Type IS FTYPE_EVENT) {
               Buffer += StrCopy(" <tr><td><a href=\"#", Buffer, COPY_ALL);
               Buffer += StrCopy(Self->FunctionArray[i]->Name, Buffer, COPY_ALL);
               Buffer += StrCopy("\">", Buffer, COPY_ALL);
               Buffer += StrCopy(Self->FunctionArray[i]->Name, Buffer, COPY_ALL);
               Buffer += StrCopy("</a>&nbsp;&nbsp;</td><td>", Buffer, COPY_ALL);
               Buffer += StrCopy(Self->FunctionArray[i]->Short, Buffer, COPY_ALL);
               Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
            }
            pos++;
            i = -1;
         }
      }
      Buffer += StrCopy("</td></tr></table></blockquote>\n\n", Buffer, COPY_ALL);
   }

   /*** Field Index ***/

   if (Self->TotalFields > 0) {
      Buffer += StrCopy("<h2>Structure</h2>\n\n<p>The ", Buffer, COPY_ALL);
      Buffer += StrCopy(Self->MH_Name, Buffer, COPY_ALL);
      Buffer += StrCopy(" object consists of the following public fields:</p>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy("<blockquote>\n<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">\n", Buffer, COPY_ALL);

      pos = 1;
      i = 0;
      while (i < Self->TotalFields) {
         if (Self->Fields[i]->Number IS pos) {
            Buffer += StrCopy(" <tr><td><a href=\"#", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Name, Buffer, COPY_ALL);
            Buffer += StrCopy("\">", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Name, Buffer, COPY_ALL);
            Buffer += StrCopy("</a>&nbsp;&nbsp;</td><td>", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Short, Buffer, COPY_ALL);
            Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
            pos++;
            i = -1;
         }
         i++;
      }
      Buffer += StrCopy("</table>\n</blockquote>\n\n", Buffer, COPY_ALL);
   }

   return(Buffer);
}

/*****************************************************************************
** Function: WriteHTMLFunctionTable()
*/

static STRING WriteHTMLFunctionTable(struct DocCompiler *Self, STRING Buffer)
{
   LONG pos, i;

   if ((Self->DocType IS DOC_MODULE) AND (Self->TotalFunctions > 0)) {
      Buffer += StrCopy("<table border=\"0\" width=\"100%\">\n", Buffer, COPY_ALL);
      Buffer += StrCopy("<tr><td colspan=\"2\"><a name=\"Functions\"></a><b><u><font face=\"Arial,Helvetica\">Function Index</font></u></b></td></tr>\n", Buffer, COPY_ALL);

      pos = 1;
      i = 0;
      while (i < Self->TotalFunctions) {
         if (Self->FunctionArray[i]->Number IS pos) {
            Buffer += StrFormat(Buffer, 1024, "<tr><td width=\"10%%\"><a href=\"#%s()\">%s</a>&nbsp;&nbsp;</td><td>%s</td></tr>\n", Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Name, Self->FunctionArray[i]->Short);
            pos++;
            i = -1;
         }
         i++;
      }
      Buffer += StrCopy("</table>\n<p>&nbsp;</p>\n\n", Buffer, COPY_ALL);
   }
   return(Buffer);
}

/*****************************************************************************
** Function: WriteHTMLFunctions()
*/

static STRING WriteHTMLFunctions(struct DocCompiler *Self, STRING Buffer)
{
   WORD order[] = { FTYPE_FUNCTION, FTYPE_ACTION, FTYPE_METHOD, FTYPE_EVENT, NULL };
   LONG pos, i, j;

   for (j=0; order[j]; j++) {
      pos = 1;
      for (i=0; i < Self->TotalFunctions; i++) {
         if (Self->FunctionArray[i]->Number IS pos) {
            if ((Self->FunctionArray[i]->Description) AND
                (Self->FunctionArray[i]->Type IS order[j])) {
               Buffer = WriteFunction(Self, Self->FunctionArray[i], Buffer);
            }
            pos++;
            i = -1;
         }
      }
   }

   return(Buffer);
}

/****************************************************************************/

STRING WriteFunction(struct DocCompiler *Self, struct DocFunction *Function, STRING Buffer)
{
   STRING header, entry;
   LONG j, k;

   Buffer += StrCopy("<table border=\"1\" cellspacing=\"0\" cellpadding=\"3\" width=\"100%\" bordercolordark=\"#003360\" bordercolorlight=\"#003360\">\n", Buffer, COPY_ALL);

   /*** Function/Method Name ***/
   if (Function->Type IS FTYPE_FUNCTION) {
      Buffer += StrCopy("<tr><td class=\"functionheader\" width=\"10%\">Function:</td><td class=\"functionentry\"><a name=\"", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      Buffer += StrCopy("()\"></a>", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      if (Function->Version > 0) {
         Buffer += StrCopy("() [V", Buffer, COPY_ALL);
         Buffer += FloatToStr(Function->Version, Buffer, COPY_ALL);
         Buffer += StrCopy("]</td></tr>\n", Buffer, COPY_ALL);
      }
      else Buffer += StrCopy("()</td></tr>\n", Buffer, COPY_ALL);
      header = "functionheader";
      entry = "functionentry";
   }
   else if (Function->Type IS FTYPE_METHOD) {
      Buffer += StrCopy("<tr><td class=\"methodheader\" width=\"10%\">Method:</td><td class=\"methodentry\"><a name=\"", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      Buffer += StrCopy("()\"></a>", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      if (Function->Version > 0) {
         Buffer += StrCopy("() [V", Buffer, COPY_ALL);
         Buffer += FloatToStr(Function->Version, Buffer, COPY_ALL);
         Buffer += StrCopy("]</td></tr>\n", Buffer, COPY_ALL);
      }
      else Buffer += StrCopy("()</td></tr>\n", Buffer, COPY_ALL);
      header = "methodheader";
      entry = "methodentry";
   }
   else if (Function->Type IS FTYPE_EVENT) {
      Buffer += StrCopy("<tr><td class=\"eventheader\" width=\"10%\">Event:</td><td class=\"evententry\"><a name=\"", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      Buffer += StrCopy("\"></a>", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      if (Function->Version > 0) {
         Buffer += StrFormat(Buffer, 1024, " [V%.1f]</td></tr>\n", Function->Version);
      }
      else Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
      header = "eventheader";
      entry = "evententry";
   }
   else /*if (Function->Type IS FTYPE_ACTION)*/ {
      Buffer += StrCopy("<tr><td class=\"actionheader\" width=\"10%\">Action:</td><td class=\"actionentry\"><a name=\"", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      Buffer += StrCopy("\"></a>", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Name, Buffer, COPY_ALL);
      if (Function->Version > 0) {
         Buffer += StrCopy(" [V", Buffer, COPY_ALL);
         Buffer += FloatToStr(Function->Version, Buffer, COPY_ALL);
         Buffer += StrCopy("]</td></tr>\n", Buffer, COPY_ALL);
      }
      else Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
      header = "actionheader";
      entry = "actionentry";
   }

   /*** Synonym ***/

   if (Function->Synonym[0]) {
      Buffer += StrFormat(Buffer, 1024, "<tr><td class=\"%s\" width=\"10%%\">Synonym:</td><td class=\"%s\">%s</td></tr>\n", header, entry, Function->Synonym);
   }

   /*** Short Description ***/

   if (Function->Short[0]) {
      Buffer += StrFormat(Buffer, 1024, "<tr><td class=\"%s\" width=\"10%%\">Short:</td><td class=\"%s\">%s</td></tr>\n", header, entry, Function->Short);
   }

   /*** Synopsis ***/

   if (Function->Synopsis[0]) {
      Buffer += StrFormat(Buffer, 1024, "<tr><td class=\"%s\" width=\"10%%\">Synopsis:</td><td class=\"%s\">%s</td></tr>\n", header, entry, Function->Synopsis);
   }

   /*** Arguments ***/
   if (Function->TotalInput > 0) {
      Buffer += StrFormat(Buffer, 1024, "<tr><td class=\"%s\" width=\"10%%\">Arguments:</td><td class=\"%s\">\n<table width=\"100%%\" border=\"1\" cellspacing=\"0\" cellpadding=\"2\" bordercolor=\"#CCCCCC\">\n", header, entry);
      for (j=0; j < Function->TotalInput; j++) {
         Buffer += StrCopy("  <tr><td width=\"10%\">", Buffer, COPY_ALL);

         for (k=0; Function->Input[j].Argument[k]; k++) {
            if (Function->Input[j].Argument[k] IS '/') {
               Buffer += StrCopy("&nbsp;", Buffer, COPY_ALL);
            }
            else *Buffer++ = Function->Input[j].Argument[k];
         }

         Buffer += StrCopy("&nbsp;&nbsp;</td><td width=\"90%\">", Buffer, COPY_ALL);
         Buffer += StrCopy(Function->Input[j].Description, Buffer, COPY_ALL);
         Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
      }
      Buffer += StrCopy("</table></td></tr>\n", Buffer, COPY_ALL);
   }

   /*** Full Description ***/

   if (Function->Description) {
      Buffer += StrCopy("<tr><td colspan=\"2\">\n", Buffer, COPY_ALL);
      Buffer += StrCopy("<table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"12\"><tr><td>\n\n", Buffer, COPY_ALL);
      Buffer += StrCopy(Function->Description, Buffer, COPY_ALL);

      /*** Results ***/

      if (Function->Results) {
         if (Function->Results->ErrorCode[0] IS NULL) {
            Buffer += StrCopy("<table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"2\">\n", Buffer, COPY_ALL);
            Buffer += StrCopy("<tr><td colspan=\"2\"><p><font face=\"Arial, Helvetica, sans-serif\"><b><u>Result</u></b></font><br>\n", Buffer, COPY_ALL);
            if (StrCompare("<p>", Function->Results->Description, 3, NULL) IS ERR_Okay) {
               Buffer += StrCopy(Function->Results->Description+3, Buffer, COPY_ALL);
            }
            else {
               Buffer += StrCopy(Function->Results->Description, Buffer, COPY_ALL);
               Buffer += StrCopy("</p>", Buffer, COPY_ALL);
            }
            Buffer += StrCopy("</td></tr></table>\n\n", Buffer, COPY_ALL);
         }
         else {
            Buffer += StrCopy("<table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"2\" bordercolor=\"#CCCCCC\">\n", Buffer, COPY_ALL);
            Buffer += StrCopy("<tr><td colspan=\"2\" class=\"tableheader\"><u>Result</u></td></tr>\n", Buffer, COPY_ALL);
            for (j=0; j < Function->TotalResults; j++) {
               Buffer += StrFormat(Buffer, 1024, "  <tr><td width=\"10%%\">%s&nbsp;&nbsp;</td><td width=\"90%%\">%s</td></tr>\n", Function->Results[j].ErrorCode, Function->Results[j].Description);
            }
            Buffer += StrCopy("</table>\n\n", Buffer, COPY_ALL);
         }
      }
      Buffer += StrCopy("</td></tr></table>\n", Buffer, COPY_ALL);
   }

   Buffer += StrCopy("</td></tr></table>\n\n<br>\n\n", Buffer, COPY_ALL);

   return(Buffer);
}

/*****************************************************************************
** Function: WriteHTMLFields()
*/

static STRING WriteHTMLFields(struct DocCompiler *Self, STRING Buffer)
{
   LONG pos, i;

   pos = 1;
   i = 0;
   while (i < Self->TotalFields) {
      if (Self->Fields[i]->Number IS pos) {
         Buffer += StrCopy("<table border=\"1\" cellspacing=\"0\" cellpadding=\"3\" width=\"100%\" bordercolordark=\"#003360\" bordercolorlight=\"#003360\">\n", Buffer, COPY_ALL);

         /*** Field Name ***/

         Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Field:</td><td class=\"fe\" width=\"90%\"><a name=\"", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->Fields[i]->Name, Buffer, COPY_ALL);
         Buffer += StrCopy("\"></a>", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->Fields[i]->Name, Buffer, COPY_ALL);
         if (Self->Fields[i]->Version > 0) {
            Buffer += StrCopy(" [V", Buffer, COPY_ALL);
            Buffer += FloatToStr(Self->Fields[i]->Version, Buffer, NULL);
            Buffer += StrCopy("]</td></tr>\n", Buffer, COPY_ALL);
         }
         else Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

         /*** Synonyms ***/

         if (Self->Fields[i]->Synonyms[0]) {
            Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Synonyms:</td><td class=\"fe\" width=\"90%\">", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Synonyms, Buffer, COPY_ALL);
            Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
         }

         /*** Short ***/

         Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Short:</td><td class=\"fe\" width=\"90%\">", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->Fields[i]->Short, Buffer, COPY_ALL);
         Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

         /*** Type ***/

         Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Type:</td><td class=\"fe\" width=\"90%\">", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->Fields[i]->Type, Buffer, COPY_ALL);
         Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

         /*** Prefix ***/

         if (Self->Fields[i]->Prefix[0]) {
            Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Prefix:</td><td class=\"fe\" width=\"90%\">", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Prefix, Buffer, COPY_ALL);
            Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);
         }

         /*** Status ***/

         Buffer += StrCopy(" <tr><td class=\"fh\" width=\"10%\">Status:</td><td class=\"fe\" width=\"90%\">", Buffer, COPY_ALL);
         Buffer += StrCopy(Self->Fields[i]->Permissions, Buffer, COPY_ALL);
         Buffer += StrCopy("</td></tr>\n", Buffer, COPY_ALL);

         /*** Full Description ***/

         if (Self->Fields[i]->Description) {
            Buffer += StrCopy("<tr><td colspan=\"2\">\n", Buffer, COPY_ALL);
            Buffer += StrCopy("<table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"12\"><tr><td>\n\n", Buffer, COPY_ALL);
            Buffer += StrCopy(Self->Fields[i]->Description, Buffer, COPY_ALL);
            Buffer += StrCopy("</td></tr></table>\n", Buffer, COPY_ALL);
         }

         Buffer += StrCopy("</td></tr></table>\n\n<br>\n\n", Buffer, COPY_ALL);

         pos++;
         i = -1;
      }

      i++;
   }

   return(Buffer);
}

/*****************************************************************************
** Function: WriteHTMLTail()
*/

static STRING WriteHTMLTail(struct DocCompiler *Self, STRING Buffer)
{
   Buffer += StrCopy("</body>\n</html>\n", Buffer, COPY_ALL);
   return(Buffer);
}

/*****************************************************************************
** Function: OutputText()
** Short:    Sends a text message to the output object.
*/

static void OutputText(struct DocCompiler *Self, STRING Text)
{
   struct acDataChannel Channel;

   if (Self->OutputID) {
      ActionMsg(AC_Clear, Self->OutputID, NULL);

      Channel.ObjectID     = GetUniqueID(Self);
      Channel.DataType       = DATA_TEXT;
      Channel.Version      = 1;
      Channel.Buffer       = Text;
      for (Channel.Size=0; Text[Channel.Size]; Channel.Size++);
      Channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &Channel);
   }
}

