/*****************************************************************************

This file is in the public domain and may be distributed and modified without
restriction.

*****************************************************************************/

#include <stdio.h>

#ifdef __linux__
#include <dlfcn.h>
#endif

#include <pandora/system/all.h>
#include <pandora/functions/strcopy.c>
#include <pandora/functions/strtoint.c>

#ifdef _WIN32
#define DLLCALL /* __declspec(dllimport) */
#define WINAPI  __stdcall
#define HKEY_LOCAL_MACHINE 0x80000002
#define KEY_READ 0x20019
DLLCALL APTR WINAPI LoadLibraryA(STRING);
DLLCALL LONG WINAPI FreeLibrary(APTR);
DLLCALL APTR WINAPI GetProcAddress(APTR, STRING);
DLLCALL LONG WINAPI RegOpenKeyExA(LONG,STRING,LONG,LONG,APTR *);
DLLCALL LONG WINAPI RegQueryValueExA(APTR,STRING,LONG *,LONG *,BYTE *,LONG *);
DLLCALL void WINAPI CloseHandle(APTR);
DLLCALL int  WINAPI MessageBoxA(LONG,STRING,STRING,LONG);
#define MB_ICONERROR 16
#define MB_OK 0
#endif

extern void program(void);
extern STRING ProgCopyright;
extern STRING ProgAuthor;
extern STRING ProgName;
extern STRING ProgDate;
extern LONG ProgDebug;
extern FLOAT ProgKernelVersion;

struct KernelBase *KernelBase;
static OBJECTID glOutputID = 0;

static ERROR PROGRAM_DataChannel(OBJECTPTR, struct acDataChannel *);
void ClosePandora(void);

/*****************************************************************************
** Function: usererror()
*/

void usererror(STRING Message)
{
#ifdef _WIN32
   MessageBoxA(NULL, Message, "Startup Failed", MB_OK|MB_ICONERROR);
#else
   printf("%s\n", Message);
#endif
}

/*****************************************************************************
** Main execution point.
*/

APTR kernelhandle = 0;
void (*closekernel)(void);
#ifdef _WIN32
APTR keyhandle;
#endif

int InitPandora(int argc, STRING *argv)
{
   #define MAX_ARGS 30
   struct OpenInfo info;
   struct KernelBase * (*openkernel)(struct OpenInfo *);
   STRING args[MAX_ARGS];
   OBJECTPTR task;
   LONG i, j, errno;
   APTR *actions;
#ifdef _WIN32
   #define SIZE_BUFFER 256
   BYTE buffer[SIZE_BUFFER];
#endif

   kernelhandle = NULL;
   openkernel   = NULL;
   closekernel  = NULL;
   errno = 1;

#ifdef __linux__

   if (!(kernelhandle = dlopen("/usr/local/pandora/system/modules/kernel.so", RTLD_NOW))) {
      kernelhandle = dlopen("/usr/local/athene/system/modules/kernel.so", RTLD_NOW);
   }

   if (!kernelhandle) {
      printf("Could not open /usr/local/pandora/system/modules/kernel.so.\n%s\n", dlerror());
      goto exit;
   }

   if (!(openkernel = dlsym(kernelhandle,"OpenKernel"))) {
      printf("\033[1mCould not find the openkernel() symbol.\033[0m\n");
      goto exit;
   }

   if (!(closekernel = dlsym(kernelhandle,"CloseKernel"))) {
      printf("\033[1mCould not find the closekernel() symbol.\033[0m\n");
      goto exit;
   }

#elif _WIN32

   keyhandle = NULL;
   j = SIZE_BUFFER;
   if (!RegOpenKeyExA(HKEY_LOCAL_MACHINE, "Software\\Rocklyte\\Pandora", 0, KEY_READ, &keyhandle)) {
      if (!RegQueryValueExA(keyhandle, "Location", 0, 0, buffer, &j)) {
         CloseHandle(keyhandle);

         for (i=0; buffer[i]; i++);
         StrCopy("system\\modules\\kernel.dll", buffer+i, SIZE_BUFFER-i-1);

         if (!(kernelhandle = LoadLibraryA(buffer))) {
            usererror("Failed to open the Pandora Engine's object kernel library.");
            goto exit;
         }
      }
      else {
         usererror("Failed to query key HKEY_LOCAL_MACHINE\\Software\\Rocklyte\\Pandora\\Location");
         goto exit;
      }
   }
   else {
      usererror("Before you can run this program you need to install the latest\nversion of the Pandora Engine.  Visit www.rocklyte.com/downloads.html\nto download this product.");
      goto exit;
   }

   if (!(openkernel = GetProcAddress(kernelhandle, "OpenKernel"))) {
      usererror("Could not find the OpenKernel() symbol in Pandora.");
      goto exit;
   }

   if (!(closekernel = GetProcAddress(kernelhandle, "CloseKernel"))) {
      usererror("Could not find the CloseKernel() symbol in Pandora.");
      goto exit;
   }

#endif

   info.Detail    = ProgDebug;
   info.MaxStep   = 10;
   info.Name      = ProgName;
   info.Author    = ProgAuthor;
   info.Date      = ProgDate;
   info.Copyright = ProgCopyright;
   info.Args      = argv;
   info.ArgCount  = argc;
   info.KernelVersion   = ProgKernelVersion; /* Minimum required kernel version */
   info.CompiledAgainst = VER_KERNEL; /* The kernel that this code is compiled against */
   info.Error           = ERR_Okay;
   info.Flags = OPF_KERNELVERSION|OPF_COMPILEDAGAINST|OPF_NAME|OPF_AUTHOR|OPF_DATE|OPF_COPYRIGHT|OPF_ARGS|OPF_ERROR;
   if (ProgDebug > 0) info.Flags |= OPF_DETAIL|OPF_MAXSTEP;
   if (ProgDebug IS -1) {
      info.Detail = 0;
      info.MaxStep = 0;
      info.Flags |= OPF_DETAIL|OPF_MAXSTEP;
   }

   if ((KernelBase = openkernel(&info))) {
      task = CurrentTask();

      GetField(task, FID_Output, FT_LONG, &glOutputID);

      if (GetField(task, FID_Actions, FT_POINTER, &actions) IS ERR_Okay) {
         actions[AC_DataChannel] = PROGRAM_DataChannel;
      }

      /*** Copy the command-line arguments into a null-terminated array ***/

      if (argc > 1) {
         j = 0;
         for (i=1; (i < argc) AND (j < MAX_ARGS-1); i++) {
            args[j++] = argv[i];
         }
         args[j++] = 0;
         SetField(task, FID_ArgsList, FT_POINTER, args);
      }

      errno = 0;
   }
   else {
      if (info.Error IS ERR_KernelVersion) usererror("This program requires the latest version of the Pandora Engine.\nPlease visit www.rocklyte.com/downloads.html to upgrade.");
      else usererror("Failed to initialise the Pandora Engine object kernel.");
   }

   return(errno);

exit:
   ClosePandora();
   return(errno);
}

/*****************************************************************************
** Function: ClosePandora()
*/

void ClosePandora(void)
{
   closekernel();

exit:
#ifdef __linux__
   if (kernelhandle) dlclose(kernelhandle);
#elif _WIN32
   if (kernelhandle) FreeLibrary(kernelhandle);
   if (keyhandle) CloseHandle(keyhandle);
#endif
}

/*****************************************************************************
** Function: print()
*/

void print(STRING text, ...)
{
   #undef SIZE_BUFFER
   #define SIZE_BUFFER 1024
   struct acDataChannel channel;
   BYTE buffer[SIZE_BUFFER];
   LONG *array, i;

   array = ((LONG *)&text)+1;

   if ((glOutputID) AND (CheckObjectExists(glOutputID, NULL) IS ERR_True)) {
      #ifdef _WIN32
         _snprintf(buffer, SIZE_BUFFER, text, array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7], array[8]);
      #else
         snprintf(buffer, SIZE_BUFFER, text, array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7], array[8]);
      #endif
      channel.ObjectID     = CurrentTaskID();
      channel.DataType     = DATA_TEXT;
      channel.Version      = 1;
      channel.Buffer       = buffer;
      for (channel.Size=0; ((STRING)channel.Buffer)[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, glOutputID, &channel);
   }
   else {
      printf(text, array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7], array[8]);
      for (i=0; text[i]; i++);
      if (text[i-1] != '\n') printf("\n");
   }
}

/*****************************************************************************
** Program: DataChannel
*/

static ERROR PROGRAM_DataChannel(OBJECTPTR Task, struct acDataChannel *Args)
{
   LONG i;

   if (Args->DataType IS DATA_TEXT) {
      UBYTE buffer[Args->Size+1];

      for (i=0; i < Args->Size; i++) buffer[i] = ((UBYTE *)Args->Buffer)[i];
      buffer[i] = 0;

      if (Args->Version <= 1) printf("%s\n", buffer);
      else if (Args->Version IS 2) printf("%s", buffer);
   }

   return(ERR_Okay);
}
